{$ASMMODE INTEL}

unit floppy;

interface

uses pit, console;

var
	Waiting : boolean;
	MotorState : byte = 0;
	Size : LongWord;
	Buffer: array[0..5119] of char;

const
	DOReg : word = $0002;
	CCReg : word = $0007; // Write-only
	MSReg : word = $0004; // Read Only - Main Status Reg
	FIFOReg : word = $0005; // Write-only
	DriveBase : array[0..1] of word = (
		$03F0, //fda
		$0370); //fdb

	FloppyCmd : array[1..19] of byte = (
		$02, // READ_TRACK          1
		$03, // SPECIFY             2
		$04, // SENSE_DRIVE_STATUS  3
		$05, // WRITE_DATA          4
		$06, // READ_DATA           5
		$07, // RECALIBRATE         6
		$08, // SENSE_INTERRUPT     7
		$09, // WRITE_DELETED_DATA  8
		$0A, // READ_ID             9
		$0C, // READ_DELETED_DATA   10
		$0D, // FORMAT_TRACK        11
		$0F, // SEEK                12
		$10, // VERSION             13
		$11, // SCAN_EQUAL          14
		$12, // PERPENDICULAR_MODE  15
		$13, // CONFIGURE           16
		$16, // VERIFY              17
		$19, // SCAN_LOW_OR_EQUAL   18
		$1D);// SCAN_HIGH_OR_EQUAL  19

	DriveStatus : array[0..3] of pchar = (
		'0','error','invalid','drive');

	MotorOnDelay : integer = {50}0;
	MotorOffDelay : integer = {300}0;
	DMABufLen : word = $4800;

procedure FloppyHandler;
function FloppyReset(Drive : byte) : byte;
function FloppyRead(Drive : byte; Data : integer; NoSec : byte) : byte;
function FloppyWrite(Drive : byte; Data : integer; NoSec : byte) : byte;
function FloppyDoTrackLBA(Drive : byte; LBA : integer; dir : byte; NoSectors : byte) : byte;
function FloppyDoTrack(Drive : byte; cyl : byte; dir : byte) : byte;
procedure LBA2CHS(LBA:integer; var cyl,head,sec : byte);
procedure FloppyWriteCmd(Drive : byte; cmd : byte);
function FloppyReadCmd(Drive : byte) : byte;
procedure IRQwait;
procedure SenseInt(Drive : byte; var st0, cyl : byte);
function Calibrate(Drive : byte) : byte;
procedure IODelay;
procedure Motor(Drive : byte; State : byte);
function Seek(Drive : byte; cyli, head : byte):byte;
procedure FloppyDMAInit(dir : byte; Size : LongWord);


implementation

procedure FloppyHandler;[public,alias:'FloppyHandler'];
begin
	 Waiting := false;
end;

function FloppyReset(Drive : byte) : byte;
var
	base : word;
	st0, cyl : byte;
begin
	st0 := -1;
	cyl := -1;

	Waiting := True;

	base := DriveBase[Drive];

	outPortb(base + DOReg, $00);
	outPortb(base + DOReg, $0C);

	IrqWait;
	SenseInt(Drive, st0, cyl);


	outPortb(base + CCReg, $00);

	FloppyWriteCmd(Drive, FloppyCmd[2]);
	FloppyWriteCmd(Drive, $DF);
	FloppyWriteCmd(Drive, $02);

	if Calibrate(Drive) = 1 then
		FloppyReset := 1
	else
		FloppyReset := 0;
end;

procedure IRQwait;
begin
	while Waiting = true do {nic};
end;

procedure SenseInt(Drive : byte; var st0, cyl : byte);
var
	base : word;
begin
	base := DriveBase[Drive];

	FloppyWriteCmd(Drive, FloppyCmd[7]);
	IODelay;

	st0 := FloppyReadCmd(Drive);
	cyl := FloppyReadCmd(Drive);
end;

procedure FloppyWriteCmd(Drive : byte; cmd : byte);
var
	base : word;
	i : integer;
begin
	base := DriveBase[Drive];

	for i := 1 to 300 do
	begin
		if (($C0 and inPortb(base + MSReg)) = $80) then
		begin
			outPortb((Base + FIFOReg), cmd);
			Exit;
		end;
		Delay(1);
	end;
end;

function Calibrate(Drive : byte) : byte;
var
	i, st0, cyl : byte ;
begin
	st0 := -1;
	cyl := -1;


	Motor(Drive, 1);

	for i:= 0 to 10 do
	begin
		Waiting := True;
		FloppyWriteCmd(Drive, FloppyCmd[6]);
		FloppyWriteCmd(Drive, Drive);
		IrqWait;
		SenseInt(Drive, st0, cyl);

		if cyl = 0 then
		begin
			Motor(Drive, 0);
			Calibrate := 0;
			Exit;
		end;
	end;
	Motor(Drive, 0);
	Calibrate := 1;
end;

procedure IODelay;
var
	i : integer;
begin
	for i := 1 to 500 do ;
end;

function FloppyReadCmd(Drive : byte) : byte;
var
	base : word;
	i : integer;
begin
	base := DriveBase[Drive];


	for i := 1 to 300 do
	begin
		if (($80 and inPortb(base + MSReg)) = $80) then
		begin
			FloppyReadCmd := inPortb(base + FIFOReg);
			Exit;
		end;
		Delay(1);
	end;
	FloppyReadCmd := 0;
end;

procedure Motor(Drive : byte; State : byte);
var
	base : word;
begin
	base := DriveBase[Drive];
// 0 - Off; 1 - On; 2 - Wait

	if State = 1 then
	begin
// If requested on
		if MotorState = 0 then
		begin
    //Turn motor on
			outPortb(base + DOReg, $1C);
			Delay(MotorOnDelay);
			MotorState := 1;
		end
		else
		begin
			Ticks := 300;
			MotorState := 2;
		end;
	end;

	if State = 0 then
		if MotorState = 1 then
		begin
			outPortb(base + DOReg, $0C);
			MotorState := 0;
			Delay(MotorOffDelay);
		end;
end;

function FloppyRead(Drive : byte; Data : integer; NoSec : byte) : byte;
begin
if Data = 0 then
	FloppyRead := FloppyDoTrack(Drive, Data, 1)
else
	FloppyRead := FloppyDoTrackLBA(Drive, Data, 1, NoSec);
end;

function FloppyDoTrackLBA(Drive : byte; LBA : integer; dir : byte; NoSectors : byte) : byte;
var
	cmd, i, st0, st1, st2, rcy, rhe, rse, bps, error : byte;
	cyl, head, sec : byte;
begin
	error := 0;

	case dir of
		1  : cmd := FloppyCmd[5] or $C0; //Read
		2  : cmd := FloppyCmd[4] or $C0; //Write
	end;

	LBA2CHS(LBA, cyl, head, sec);
	Size := NoSectors * 512;

	if (Seek(Drive, cyl, head)) = -1 then
	begin
		FloppyDoTrackLBA := -1;
		Exit;
	end;

//	for i := 1 to 20 do
	begin
		Motor(Drive, 1);
		FloppyDMAInit(dir, Size);

		Delay(10);

		FloppyWriteCmd(Drive, cmd);
		FloppyWriteCmd(Drive, ((head shl 2) or Drive));
		FloppyWriteCmd(Drive, cyl);
		FloppyWriteCmd(Drive, head);
		FloppyWriteCmd(Drive, sec);
		FloppyWriteCmd(Drive, 2);
		FloppyWriteCmd(Drive, 0);
		FloppyWriteCmd(Drive, 0);
		FloppyWriteCmd(Drive, $FF);

		IRQWait;

		st0 := FloppyReadCmd(Drive);
		st1 := FloppyReadCmd(Drive);
		st2 := FloppyReadCmd(Drive);

		//Read Cylinder, Head, Sector
		rcy := FloppyReadCmd(Drive);
		rhe := FloppyReadCmd(Drive);
		rse := FloppyReadCmd(Drive);

		bps := FloppyReadCmd(Drive);

		if error = 0  then
		begin
			Motor(Drive, 0);
			FloppyDoTrackLBA := 0;
			Exit;
		end;

		if error >= 1 then
		begin
			Motor(Drive, 0);
			FloppyDoTrackLBA := -2;
			Exit;
		end;
	end;
 
	Motor(Drive, 0);
	FloppyDoTrackLBA := -1;
end;

procedure LBA2CHS(LBA:integer; var cyl,head,sec : byte);
begin
	sec := (LBA mod 18) + 1;
	cyl := LBA Div 36;
	head := (LBA Div 18) mod 2;
end;  

function Seek(Drive : byte; cyli, head : byte):byte;
var
	st0, cyl, i : byte;
begin

	st0 := -1;
	cyl := -1;

	Motor(Drive, 1);

	for i := 0 to 10 do
	begin
		Waiting := True;
		FloppyWriteCmd(Drive, FloppyCmd[12]);
		FloppyWriteCmd(Drive, (head shl 2));
		FloppyWriteCmd(Drive, cyli);
		IRQWait;
		SenseInt(Drive, st0, cyl);

		if cyl = cyli then
		begin
			Motor(Drive,0);
			Seek := 0;
			Exit;
		end;
	end;
	Seek := -1;
	Motor(Drive, 0);
end;

procedure FloppyDMAInit(dir : byte; Size : LongWord);
var
	Addr, Count : Longword;
	mode : byte;
begin
	Addr := DWord(@Buffer);
	Count := Size - 1;

	mode := $40;
	mode := mode or $02;

	case dir of
		1  : mode := mode or $04; //Read to mem
		2  : mode := mode or $08; //Write from mem
	end;

	outPortb($0A,$06); //Mask DMA 2
	outPortb($0C,$FF); //Reset Flip-Flop

	outPortb($04,(Addr and $000000FF));         //Address Low byte
	outPortb($04,((Addr shr 8) and $000000FF)); //Address High byte

	outPortb($81,((Addr shr 16) and $000000FF)); //External Page Register

	outPortb($0C,$FF); //Reset Flip-Flop
	outPortb($05,(Count and $000000FF));         //Count low byte
	outPortb($05,((Count shr 8) and $000000FF)); //Count High byte

	outPortb($0B,Mode); //Set Mode -see above
	outPortb($0A,$02);  //Unmask DMA 2
end;

function FloppyWrite(Drive : byte; Data : integer; NoSec : byte) : byte;
begin
	FloppyWrite := FloppyDoTrackLBA(Drive, Data, 2, NoSec);
end;

function FloppyDoTrack(Drive : byte; cyl : byte; dir : byte) : byte;
var
cmd, i, st0, st1, st2, rcy, rhe, rse, bps, error : byte;
begin
error := 0;

case dir of
1  : cmd := FloppyCmd[5] or $C0; //Read
2  : cmd := FloppyCmd[4] or $C0; //Write
end;

Size := DMABufLen;

if (Seek(Drive, cyl, 0)) or (Seek(Drive, cyl, 1)) = -1 then
  begin
  FloppyDoTrack := -1;
  Exit;
  end;

//for i := 1 to 20 do
 begin
   Motor(Drive, 1);
   FloppyDMAInit(dir, Size);

   Delay(10);

   FloppyWriteCmd(Drive, cmd);
   FloppyWriteCmd(Drive, 0);
   FloppyWriteCmd(Drive, cyl);
   FloppyWriteCmd(Drive, 0);
   FloppyWriteCmd(Drive, 1);
   FloppyWriteCmd(Drive, 2);
   FloppyWriteCmd(Drive, 18);
   FloppyWriteCmd(Drive, $1B);
   FloppyWriteCmd(Drive, $FF);

   IRQWait;

   st0 := FloppyReadCmd(Drive);
   st1 := FloppyReadCmd(Drive);
   st2 := FloppyReadCmd(Drive);

   //Read Cylinder, Head, Sector
   rcy := FloppyReadCmd(Drive);
   rhe := FloppyReadCmd(Drive);
   rse := FloppyReadCmd(Drive);

   bps := FloppyReadCmd(Drive);

   if error = 0  then
    begin
      Motor(Drive, 0);
      FloppyDoTrack := 0;
      Exit;
    end;

   if error >= 1 then
    begin
      Motor(Drive, 0);
      FloppyDoTrack := -2;
      Exit;
    end;
 end;
 
Motor(Drive, 0);
FloppyDoTrack := -1;
end;

end. 